/**
 * File: button.c
 * The logic for interpretting the button presses into several events (single vs double vs triple vs long presses) is implemented here 
 * Once the event is interpretted, this module calls the LED module to take appropriate action for the given event.
*/

#include <string.h>

#include "gpio.h"
#include "pins.h"
#include "timers.h"
#include "led.h"

// Button press durations
#define MIN_GAP_BTW_PRESSES_MS 200
#define MAX_GAP_BTW_PRESSES_MS 800
#define LONG_PRESS_DURATION_MS 2000


// timer instance allocations
#define BUTTON_A_LONG_TIMER  TIMER_1
#define BUTTON_A_SHORT_TIMER TIMER_2

#define BUTTON_B_LONG_TIMER  TIMER_3
#define BUTTON_B_SHORT_TIMER TIMER_4

#define MAX_NUM_PRESSES 3

/**************************************************Enums and Structs to hold app data**************************************************************/
enum actions {
    NO_ACTION,
    BUTTON_A_LONG_PRESS,
    BUTTON_A_SINGLE_PRESS,
    BUTTON_A_DOUBLE_PRESS,
    BUTTON_A_TRIPLE_PRESS,
    BUTTON_B_LONG_PRESS,
    BUTTON_B_SINGLE_PRESS,
    BUTTON_B_DOUBLE_PRESS
} _action;

enum events {
    NO_EVENT,
    BUTTON_A_LONG_PRESS_DONE,
    BUTTON_A_LAST_PRESS_REGISTERED,
    BUTTON_B_LONG_PRESS_DONE,
    BUTTON_B_LAST_PRESS_REGISTERED
} _button_event;

typedef struct {
    int state;
    int num_clicks;
    enum timer_instance dur_pressed_timer;
    enum timer_instance dur_released_timer;
    int dur_btw_presses[MAX_NUM_PRESSES];
} button_t;

static volatile button_t _button_A;
static volatile button_t _button_B;

/**************************************************Helper Functions*****************************************************************************************/

static void _reset_button(button_t * p_button)
{
    p_button->state = BUTTON_RELEASED;
    p_button->num_clicks = 0;
    memset(p_button->dur_btw_presses, 0, sizeof(int)*MAX_NUM_PRESSES);
}

static void _take_action()
{
    switch (_action) {
        case BUTTON_A_LONG_PRESS:
            led_default_brightness();
            break;

        case BUTTON_A_SINGLE_PRESS:
            led_toggle_white();
            break;

        case BUTTON_A_DOUBLE_PRESS:
            led_step_up_brightness();
            break;

        case BUTTON_A_TRIPLE_PRESS:
            led_step_down_brightness();
            break;

        case BUTTON_B_LONG_PRESS:
            led_toggle_power();
            break;

        case BUTTON_B_SINGLE_PRESS:
            led_next_color();
            break;

        case BUTTON_B_DOUBLE_PRESS:
            led_prev_color();
            break;
        
        default:
            // do nothing
            break;
    }
}



static void _set_action()
{
    switch(_button_event) {
        case BUTTON_A_LONG_PRESS_DONE:
            _action = BUTTON_A_LONG_PRESS;
            break;

        case BUTTON_B_LONG_PRESS_DONE:
            _action = BUTTON_B_LONG_PRESS;
            break;        

        case BUTTON_A_LAST_PRESS_REGISTERED:
            if (_button_A.num_clicks == 1) {
                _action = BUTTON_A_SINGLE_PRESS;
            } else if (_button_A.num_clicks ==2) {
                _action = BUTTON_A_DOUBLE_PRESS;
            } else if (_button_A.num_clicks == 3) {
                _action = BUTTON_A_TRIPLE_PRESS;
            } else {
                _action = NO_ACTION;
            }
            _reset_button(&_button_A);
            break;        

        case BUTTON_B_LAST_PRESS_REGISTERED:
            if (_button_B.num_clicks == 1) {
                _action = BUTTON_B_SINGLE_PRESS;
            } else if (_button_B.num_clicks ==2) {
                _action = BUTTON_B_DOUBLE_PRESS;
            } else {
                _action = NO_ACTION;
            }
            _reset_button(&_button_B);
            break;

        default: 
            _action = NO_ACTION;
            break;
    }
    if(_action != NO_ACTION) {
        _take_action();
    }
}


static void _set_timer_on(enum timer_instance timer, int count)
{
    timer_interrupt_disable(timer);
    timer_stop(timer);
    timer_clear_count(timer);
    timer_set_count(timer, count);
    timer_interrupt_enable(timer);
    timer_start(timer);
}

static void _handle_short_press_event(button_t * p_button)
{
    if (p_button->state == BUTTON_PUSHED) {
        // stop the dur_released_timer and read released duration 
        timer_stop(p_button->dur_released_timer);
        // the timer counts down, hence we need to subract the present value from intial set value
        p_button->dur_btw_presses[p_button->num_clicks] = LONG_PRESS_DURATION_MS - timer_get_count(p_button->dur_released_timer);

        // for the first button press register here
        if ( p_button->num_clicks == 0) {
            p_button->num_clicks++;
        }
        
        // For subsequent presses check if the time duration between button presses the right amount of time
        // if yes then increment the num_clicks
        // Note: the condition where button is pressed again too late is handled in the timer handler 
        if (p_button->dur_btw_presses[p_button->num_clicks] >= MIN_GAP_BTW_PRESSES_MS) {
            p_button->num_clicks++;
        }
    } else {
        // start the dur_released_timer for MAX_GAP_BTW_PRESSES_MS, to measure released duration
        _set_timer_on(p_button->dur_released_timer, LONG_PRESS_DURATION_MS);
    }
}


static void _handle_long_press_event(button_t * p_button)
{
    if (p_button->state == BUTTON_PUSHED) {
        // start the dur_pressed_timer for LONG_PRESS_DURATION_MS, to measure pressed duration
        _set_timer_on(p_button->dur_pressed_timer, LONG_PRESS_DURATION_MS);
    } else {
        // stop the dur_released_timer
        timer_stop(p_button->dur_released_timer);
    }
}


static void _handle_event(button_t * p_button)
{
    _handle_short_press_event(p_button);
    _handle_long_press_event(p_button);
    _set_action();
}

/**************************************************Interrupt Handler Implementations*****************************************************************/

void gpio_handler()
{
    enum gpio_pin button = GPIO_get_interrupt_source();
    // read status of button
    int button_state =  GPIO_read_digital(button);

    // disable interrupt on that button to avoid another interrupt while this one is being processed
    GPIO_interrupt_clear(button);
    GPIO_interrupt_disable(button);

    // store the state in the main datastructure
    button_t * p_button;
    if (button == BUTTON_A) {
        p_button = &_button_A; 
    } else if (button == BUTTON_B) {
        p_button = &_button_B; 
    }
    p_button->state = button_state;
    // call the event handler
    _handle_event(p_button);
    GPIO_interrupt_enable(button);
}


void timer_handler()
{
    enum timer_instance timer = timer_get_interrupt_source();
    timer_interrupt_disable(timer);
    // if the timer expires, then the event is a long-press

    switch (timer) {
        case BUTTON_A_LONG_TIMER:
            _button_event = BUTTON_A_LONG_PRESS_DONE;
            break;

        case BUTTON_A_SHORT_TIMER:
            _button_event = BUTTON_A_LAST_PRESS_REGISTERED;
            break;

        case BUTTON_B_LONG_TIMER:
            _button_event = BUTTON_B_LONG_PRESS_DONE;
            break;

        case BUTTON_B_SHORT_TIMER:
            _button_event = BUTTON_B_LAST_PRESS_REGISTERED;
            break;

        default:
            _button_event = NO_EVENT;
            break;
    }
    _set_action();
    timer_interrupt_enable(timer);
}


/**************************************************Initialization functions***********************************************************/
static void _init_gpio()
{
    //init buttons to be DIGITAL IO 
    GPIO_set_func(BUTTON_A, DIGITAL_IO);
    GPIO_set_func(BUTTON_B, DIGITAL_IO);

    // configure the buttons as inputs
    GPIO_set_direction(BUTTON_A, INPUT);
    GPIO_set_direction(BUTTON_B, INPUT);
}


static void _init_timer_instance(enum timer_instance timer)
{
    timer_enable(timer);
    timer_set_mode(timer, ONE_SHOT);
    timer_clear_count(timer);
}


static void _init_timers()
{
    _init_timer_instance(BUTTON_A_LONG_TIMER);
    _init_timer_instance(BUTTON_A_SHORT_TIMER);
    _init_timer_instance(BUTTON_B_LONG_TIMER);
    _init_timer_instance(BUTTON_B_SHORT_TIMER);
}


static void _enable_interrupts()
{
    // clear any previous interrupts
    GPIO_interrupt_clear(BUTTON_A);
    GPIO_interrupt_clear(BUTTON_B);

    // enable interrupts
    GPIO_interrupt_enable(BUTTON_A);
    GPIO_interrupt_enable(BUTTON_B);
}

static void _init_button_structs()
{
    memset(&_button_A, 0, sizeof(_button_A));
    memset(&_button_B, 0, sizeof(_button_B));

    _button_A.dur_pressed_timer = BUTTON_A_LONG_TIMER;
    _button_A.dur_released_timer = BUTTON_A_SHORT_TIMER;
    _button_B.dur_pressed_timer = BUTTON_B_LONG_TIMER;
    _button_B.dur_released_timer = BUTTON_B_SHORT_TIMER;
}

/************************************************************Interface functions*******************************************************************/


void button_init()
{
    _init_gpio();
    _init_timers();
    _init_button_structs();
    _enable_interrupts();
}

/***********************************************************************************************************************************************/