/**
 * File: led.c
 * The actual actions of changing LED colors and brightness according to the button event is implemented here.
*/

#include <stdbool.h>
#include "gpio.h"
#include "pwm.h"
#include "pins.h"

#define LED_DEF_BRIGHTNESS_PERCENT PWM_DUTY_50_PERCENT

typedef struct
{
    bool on;
    enum duty_cycle brightness;
} led_t;

static volatile led_t _red, _green, _blue;
static volatile bool _power_on;

enum {
    WHITE_COLOR,
    RED_COLOR,
    GREEN_COLOR,
    BLUE_COLOR,
    YELLOW_COLOR
} _led_state;

/**************************************************Helper functions**********************************************************************************/

static void _turn_leds_off()
{
    PWM_set_duty_cycle(LED_RED, PWM_DUTY_0_PERCENT);
    PWM_set_duty_cycle(LED_BLUE, PWM_DUTY_0_PERCENT);
    PWM_set_duty_cycle(LED_GREEN, PWM_DUTY_0_PERCENT);
}


static void _init_brightness()
{
    _red.brightness = LED_DEF_BRIGHTNESS_PERCENT;
    _green.brightness = LED_DEF_BRIGHTNESS_PERCENT;
    _blue.brightness = LED_DEF_BRIGHTNESS_PERCENT;
}


static void _set_brightness()
{
    if (_power_on) {
        if(_red.on) {
            PWM_set_duty_cycle(LED_RED, _red.brightness);
        }
        if(_green.on) {
            PWM_set_duty_cycle(LED_GREEN, _green.brightness);
        }
        if(_blue.on) {
            PWM_set_duty_cycle(LED_BLUE, _blue.brightness);
        }
    }
}


static void _set_color()
{
    _init_brightness();
    switch (_led_state) {
        case WHITE_COLOR:
            _red.on = true;
            _green.on = true;
            _blue.on = true;
            break;
            
        case RED_COLOR:
            _red.on = true;
            _green.on = false;
            _blue.on = false;            
            break;

        case GREEN_COLOR:
            _red.on = false;
            _green.on = true;
            _blue.on = false;            
            break;

        case BLUE_COLOR:
            _red.on = false;
            _green.on = false;
            _blue.on = true;            
            break;

        case YELLOW_COLOR: // turning red and green on together gives yellow color
            _red.on = true;
            _green.on = true;
            _blue.on = false;            
            break;

        default:
            // do nothing
            break;
    }
    _set_brightness();
}


/**************************************************Initialization functions***********************************************************/
static void _init_gpio()
{
    // init LEDs to use PWM
    GPIO_set_func(LED_RED, PWM);
    GPIO_set_func(LED_BLUE, PWM);
    GPIO_set_func(LED_GREEN, PWM);
}


static void _init_pwm()
{
    // enable the PWM module on the LED pins
    PWM_enable(LED_RED);
    PWM_enable(LED_BLUE);
    PWM_enable(LED_GREEN);
}


static void _init_states()
{
    _red.on = true;
    _green.on = true;
    _blue.on = true;
    _led_state = WHITE_COLOR;
    _power_on = true;
}

/**************************************************Interface functions*******************************************************************************/

void led_init()
{
    _init_gpio();
    _init_pwm();
    _init_brightness();
    _set_brightness();
    _init_states();
}

void led_toggle_white()
{
    if (_red.on || _green.on || _blue.on) {
        _turn_leds_off();
    } else {
        _red.on = true;
        _green.on = true;
        _blue.on = true;
        _set_brightness();
    }
}

void led_step_up_brightness()
{
    _red.brightness++;
    _green.brightness++;
    _blue.brightness++;
    _set_brightness();
}

void led_step_down_brightness()
{
    _red.brightness--;
    _green.brightness--;
    _blue.brightness--;
    _set_brightness();
}

void led_default_brightness()
{
    _init_brightness();
    _set_brightness();
}


void led_next_color()
{
    if (_led_state < YELLOW_COLOR) {
        _led_state++;
    } else {
        _led_state = WHITE_COLOR;
    }
    _set_color();
}

void led_prev_color()
{
    if (_led_state > WHITE_COLOR) {
        _led_state--;
    } else {
        _led_state = YELLOW_COLOR;
    }
    _set_color();
}

void led_toggle_power()
{
    if(_power_on) {
        _power_on = false;
    } else {
        _power_on = true;
    }
    _set_brightness();
}

/***********************************************************************************************************************************************/
